#!/usr/bin/env python

import os, sys, pickle, json

import pytest
from ampel.log.AmpelLogger import AmpelLogger

logger = AmpelLogger.get_logger()

# import pickle

# with open(
#     "/Users/simeon/ampel-nuclear/ampel/nuclear/test/test_tv.pickle", "wb"
# ) as f:
#     pickle.dump(transients[0], f)


@pytest.fixture
def lightCurve_fixture():
    with open(
        os.path.join(os.path.dirname(__file__), "ZTF20aaadtjq_transientview.pickle"),
        "rb",
    ) as f:
        tv = pickle.load(f)

    print(tv)
    lightCurve = tv.get_lightcurves()[0]

    return lightCurve


@pytest.fixture
def true_output():
    truth = {}
    truth["flex"] = json.load(
        open(
            os.path.join(os.path.dirname(__file__), "ZTF20aaadtjq_flexresult.json"),
            "rb",
        )
    )
    truth["simple"] = json.load(
        open(
            os.path.join(os.path.dirname(__file__), "ZTF20aaadtjq_simpleresult.json"),
            "rb",
        )
    )
    return truth


def test_t2_flexfit(lightCurve_fixture, true_output):
    from ampel.nuclear.t2.T2FlexFit import T2FlexFit

    # test T2FlexFit
    flexT2 = T2FlexFit(logger=logger, oldest_upper_limits=14, max_post_peak=200)
    flex_result = flexT2.process(lightCurve_fixture)
    flex_truth = true_output["flex"]

    assert list(flex_result.keys()) == [
        "fit_params",
        "lightcurve_data",
        "plot_info",
    ]  # check to make sure output dict is full
    assert flex_result["fit_params"]["name"] == flex_truth["fit_params"]["name"]


def test_t2_simplemetrics(lightCurve_fixture, true_output):
    from ampel.nuclear.t2.T2SimpleMetrics import T2SimpleMetrics

    #     # test T2SimpleMetrics
    simpleT2 = T2SimpleMetrics(logger=logger)
    simple_result = simpleT2.process(lightCurve_fixture)
    simple_truth = true_output["simple"]
    assert list(simple_result.keys()) == [
        "metrics",
        "plot_info",
    ]  # Ensure proper output

    assert simple_result["metrics"]["name"] == simple_truth["metrics"]["name"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : ampel/nuclear/t3/T3Ranking.py
# License           : BSD-3-Clause
# Author            : mitchell@nyu.edu
# Date              : 08.06.2020
# Last Modified Date: 29.11.2022
# Last Modified By  : simeon.reusch@desy.de

import datetime, json, io, time, os

from typing import Any, Optional, Union
from collections.abc import Generator

import matplotlib as mpl  # type: ignore
from matplotlib import pyplot as plt  # type: ignore
import numpy as np
import pandas as pd  # type: ignore
import astropy.time  # type: ignore
import corner  # type: ignore

from ampel.types import UBson, T3Send
from ampel.view.TransientView import TransientView
from ampel.view.T3Store import T3Store
from ampel.struct.UnitResult import UnitResult
from ampel.abstract.AbsPhotoT3Unit import AbsPhotoT3Unit
from ampel.nuclear.t3.dropboxIO import DropboxUnit
import ampel.nuclear.t3.classifyme as classifyme

flex_keys = [
    "dtime_peak",
    "band",
    "mag_peak",
    "sigma_rise",
    "sigma_fade",
    "mean_color",
    "color_slope",
    "e_dtime_peak",
    "e_mag_peak",
    "e_mean_color",
    "e_color_slope",
    "e_sigma_rise",
    "mad",
    "rms",
    "chi2",
    "n_rise",
    "n_fade",
    "classification",
    "name",
    "mjd_peak",
    "photoclass",
    "blabber",
]

simple_keys = [
    "name",
    "classification",
    "ra",
    "dec",
    "ndetections",
    "ndetections_negative",
    "peak_mag",
    "latest_mag",
    "peak_jd",
    "latest_jd",
    "age",
    "ndetections_g",
    "offset_rms_g",
    "offset_unc_g",
    "offset_med_g",
    "offset_weighted_g",
    "offset_sig_g",
    "ndetections_r",
    "offset_rms_r",
    "offset_unc_r",
    "offset_med_r",
    "offset_weighted_r",
    "offset_sig_r",
    "offset_sig",
    "lastest_obs_is_detection",
    "latest_maglim",
    "ref_mag_r",
    "ref_mag_g",
    "peak_mag_g",
    "latest_mag_g",
    "peak_diff_mag_g",
    "latest_diff_mag_g",
    "peak_mag_r",
    "latest_mag_r",
    "peak_diff_mag_r",
    "latest_diff_mag_r",
    "ref_mag",
    "latest_diff_mag",
    "peak_diff_mag",
]


class T3Ranking(DropboxUnit):

    plotHisto: bool = False  # create histrograms for of the magnitude of the transients in the different summary plots

    def post_init(self):
        # super().post_init()

        # make empty dataframes
        self.metrics = pd.DataFrame(columns=simple_keys)
        self.metrics_flex = pd.DataFrame(columns=flex_keys)

        # do folder I/O
        self.ranking_location = self.base_location + "/ranking"
        self.sum_location = self.base_location + "/sum_plots"

        if "ranking" not in self.get_files(self.base_location):
            self.create_folder(self.ranking_location)

        if "sum_plots" not in self.get_files(self.base_location):
            self.create_folder(self.sum_location)

        # if needed, add the year subfolder (always needed when we are running pytest)
        self.this_year = str(self.night.datetime.year)
        if self.this_year not in self.get_files(self.ranking_location):
            self.create_folder(self.ranking_location + "/" + self.this_year)

        sum_location_year = self.sum_location + "/" + self.this_year
        if self.this_year not in self.get_files(self.sum_location):
            self.create_folder(sum_location_year)

        # finally, add the dd-mm subfolder for sum_plot
        self.this_mmdd = self.night.datetime.strftime("%m-%d")
        self.plot_dir = sum_location_year + "/" + self.this_mmdd
        if self.this_mmdd not in self.get_files(sum_location_year):
            self.create_folder(self.plot_dir)

    def process(
        self, gen: Generator[TransientView, T3Send, None], t3s: Optional[T3Store] = None
    ) -> Union[UBson, UnitResult]:
        """ """
        transients = [t for t in gen]

        metrics, metrics_flex = self.collect_metrics(transients)

        if len(metrics):
            self.metrics = pd.concat([self.metrics, metrics], ignore_index=True)
            self.metrics_flex = pd.concat(
                [self.metrics_flex, metrics_flex], ignore_index=True
            )
            self.logger.info(
                "done adding transients, total collected: " + str(len(self.metrics))
            )
            self.make_sumplots(self.metrics, self.metrics_flex)
        else:
            self.metrics = []
            self.metrics_flex = []
            self.logger.info("metrics is empty, skipping summary plots")

        if len(self.metrics):
            self.apply_ranking(self.metrics, self.metrics_flex)

        super().done()

        return None

    def collect_metrics(self, transients):
        """ """

        if transients:

            local_sources = np.array(transients)
            newdetection_count = len(local_sources)
            self.logger.info(
                "Collecting metrics for {} transient(s)".format(newdetection_count)
            )

            simple_res_list = []
            flexfit_res_list = []

            for i, tran_view in enumerate(local_sources):

                # get the classification and some extra info
                # writing of the classifcation to dropbox is done by T3SummaryPlots
                classification, extra_info = classifyme.get_class(
                    tran_view, self, self.logger, write=False
                )

                simple_res = self.get_simple_results(
                    tran_view, classification, extra_info
                )
                flexfit_res = self.get_fit_params(tran_view, classification)

                simple_res_list.append(simple_res)
                flexfit_res_list.append(flexfit_res)

            metrics = pd.DataFrame.from_records(simple_res_list)
            metrics_flex = pd.DataFrame.from_records(flexfit_res_list)

            self.logger.info(
                "successfully collected metrics: "
                + str(all(metrics_flex["name"] == metrics["name"]))
            )
        else:
            self.logger.info("no transients to rank!")
            return [], []

        # Cast object columns to float64
        for k in metrics_flex.columns:
            if k not in {
                "name",
                "classification",
                "photoclass",
                "blabber",
                "n_rise",
                "n_fade",
                "band",
            }:
                metrics_flex[k] = metrics_flex[k].astype(float)

        # Returns results from T2s
        return metrics, metrics_flex

    def get_fit_params(self, tran_view, classification):
        """ """
        try:
            fit_params = dict(
                (tran_view.get_latest_t2_body("T2FlexFit") or {})["fit_params"]
            )
            fit_params["classification"] = classification
            return fit_params

        except KeyError:
            self.logger.warn(
                "WARNING: No Flex fit for source {}".format(tran_view.stock["name"][0])
            )
            emptyrow = dict(zip(np.full(len(flex_keys), None), flex_keys))
            emptyrow["name"], emptyrow["classification"] = (
                tran_view.stock["name"][0],
                classification,
            )
            return emptyrow

    def get_simple_results(self, tran_view, classification, extra_info):
        """ """
        simple_results = dict(
            tran_view.get_latest_t2_body("T2SimpleMetrics" or {}).get("metrics", {})
        )

        simple_results["ra"] = np.mean(simple_results["ra"])
        simple_results["dec"] = np.mean(simple_results["dec"])

        simple_results["classification"] = classification
        simple_results["extra_info"] = extra_info

        # get all the jds for the lightcurve photopoints -> might provide useful
        # lc_photopoints = tran_view.get_lightcurves()[0].get_photopoints()
        # lc_jds = [p.get("body").get("jd") for p in lc_pps]

        simple_results["age"] = (
            self.night.jd
            - astropy.time.Time(tran_view.get_time_updated(output="datetime")).jd
        )

        # to do: read this from the T2 catalog match records?
        alert_base = "/mampel/alerts/" + "20" + simple_results["name"][3:5]
        try:
            # if matched to PS1, read offset and our estimate of the offset significance
            # this could also be computed from the T2 match info, but this is fast enough
            results_infile = self.read_file(
                alert_base
                + "/{}/{}_simple.json".format(
                    simple_results["name"], simple_results["name"]
                )
            )
            if isinstance(results_infile, str):
                results_json = json.loads(results_infile)
            else:
                results_json = results_infile.json()

            simple_results["offset_med_ps1"] = results_json["offset_med_ps1"]
            simple_results["offset_rms_ps1"] = results_json["offset_rms_ps1"]
            simple_results["offset_sig_ps1"] = results_json["offset_sig_ps1"]
            self.logger.info(
                "PS1 offsets calculated for {}".format(simple_results["name"])
            )
        except (FileNotFoundError, KeyError) as e:
            self.logger.warn(
                "WARNING: No PS1 offsets calculated for source {}".format(
                    tran_view.stock["name"][0]
                )
            )
            simple_results["offset_med_ps1"] = 0.0
            simple_results["offset_rms_ps1"] = 0.0
            simple_results["offset_sig_ps1"] = 0.0

        return simple_results

    def apply_ranking(self, metrics, metrics_flex):
        """ """

        ranking = self.get_ranking(metrics, metrics_flex)
        self.metrics["ranking"] = ranking
        metrics = self.metrics
        metrics_flex = self.metrics_flex

        if self.dryRunDir:
            metrics.to_csv(os.path.join(self.dryRunDir, "metrics.csv"))
            metrics_flex.to_csv(os.path.join(self.dryRunDir, "metrics_flex.csv"))

        base_name = (
            self.ranking_location + "/" + self.this_year + "/" + self.this_mmdd
        )  # base of filename

        metrics["classification"] = [
            "None" if (c == "") or (c == np.nan) else c
            for c in metrics["classification"]
        ]

        # disable for now
        iage = metrics["age"] > -9999

        # Select things saved for as part of Ampel filter
        # iampel = ['ZTFBH Nuclear' in [x['comment'] for x in aa] for aa in [ls['autoannotations'] for ls in marshal_data['metadata']]]

        iagn = (
            (metrics["classification"] == "AGN")
            | (metrics["classification"] == "blazar")
            | (metrics["classification"] == "QSO")
            | (metrics["classification"] == "Blazar")
        )  # This would use AGN detection...
        icv = np.array(["CV" in str(x) for x in metrics["classification"]])
        istar = np.array(["st" in str(x) for x in metrics["classification"]]) | icv
        ibogus = np.array(["bogus" in str(x) for x in metrics["classification"]])
        iflexok = (
            (metrics_flex["chi2"] < 10)
            & (metrics_flex["photoclass"] != "no fit")
            & (metrics_flex["photoclass"] != "fast/weird")
        )

        iok = (
            (metrics["peak_mag"] < 20)
            & (metrics["peak_diff_mag"] < 2)
            & (metrics["peak_mag"] > 0)
        )

        ipublic = (
            (metrics["latest_mag"] < 19.5)
            & (metrics["latest_diff_mag"] < 1)
            & (
                metrics["ndetections"]
                > 4 & metrics["lastest_obs_is_detection"].astype(bool)
            )
        )

        inone = metrics["classification"] == "None"  # Marshal classification
        itde = (metrics["classification"] == "TDE") | (
            metrics["classification"] == "TDE?"
        )  # Marshal classification
        imaybe = ["?" in str(x["classification"]) for ind, x in metrics.iterrows()]
        inotsecure = inone | imaybe

        # to do: get this info again from Fritz (and Growth?)
        # ihasspec = np.array([bool(ls['n_spectra']) for ls in marshal_data['metadata']])
        # self.logger.info ('# of sources with spectra: '+  str(sum(ihasspec)))
        # inospec = ihasspec==False

        # try a public stream
        iselect = (
            ipublic
            & iage
            & iflexok
            & (iagn == False)
            & (ibogus == False)
            & (istar == False)
        )
        self.logger.info("# sources in public list: " + str(sum(iselect)))
        filename = base_name + "_public_testing.txt"
        self.print_ranking(filename, metrics[iselect], metrics_flex[iselect])

        # spectrum without secure marshal class
        # iselect =  iok & iflexok & inotsecure # & ihasspec
        # filename = base_name.split('_')[0]+'_spectrum_but_no_classification.txt'
        # self.print_ranking(filename, metrics[iselect], metrics_flex[iselect])

        # everything (including bad fit etc)
        iselect = iage
        self.logger.info("# sources in everything list: " + str(sum(iselect)))
        filename = base_name + "_everything.txt"

        self.print_ranking(filename, metrics[iselect], metrics_flex[iselect])

        # SEDM select bright sources
        # to do: filter out sources wih a spectrum
        ibrightnow = metrics["latest_mag"] < 19.5
        iselect = ibrightnow & iflexok & iage & iok & inotsecure  # & inospec
        self.logger.info("# sources in SEDM list: " + str(sum(iselect)))

        filename = base_name + "_unclassified_SEDM.txt"
        self.print_ranking(filename, metrics[iselect], metrics_flex[iselect])

        # select TDE candidates, plus known TDEs
        iphototde = (metrics_flex["photoclass"] == "TDE") | (
            metrics_flex["photoclass"] == "TDE?"
        )
        iselect = iage & iok & iflexok & iphototde & (inotsecure | itde)
        self.logger.info("# sources in TDE list: " + str(sum(iselect)))

        filename = base_name + "_TDEs.txt"
        self.print_ranking(filename, metrics[iselect], metrics_flex[iselect])

        # select source for rapid Swift follow-up:
        # blue, slow and pre-peak
        ipretde = (
            (metrics_flex["mean_color"] < 0.0)
            & (np.abs(metrics_flex["e_mean_color"]) < 0.35)
            & (metrics_flex["sigma_rise"] > 9)
            & (
                metrics_flex["sigma_rise"]
                / np.clip(metrics_flex["e_sigma_rise"], 1, 1e99)
                > 2
            )
            & (metrics_flex["n_rise"] > 2)
            & (metrics_flex["n_fade"] < 3)
        )
        iselect = iage & iok & iflexok & ipretde & (inotsecure | itde)  # &inospec
        self.logger.info("# sources in pre-peak TDE list: " + str(sum(iselect)))
        filename = base_name + "_prepeak_TDEs.txt"

        self.print_ranking(filename, metrics[iselect], metrics_flex[iselect])

    def get_ranking(self, metrics, metrics_flex):
        """
        attempt at ranking sources for follow-up, using resulting from
        light curve fitting (flexfit) and basic properties  (simplemetrics)

        """

        rank = np.zeros(len(metrics), dtype=float)
        rank[
            metrics["peak_mag"] == 0
        ] = 1000  # if it didnt pass filter, even lower rank
        ipbad = np.array([bool("bad" in str(pc)) for pc in metrics_flex["photoclass"]])
        ipagn = np.array([bool("AGN" in str(pc)) for pc in metrics_flex["photoclass"]])
        ipcv = np.array([bool("CV" in str(pc)) for pc in metrics_flex["photoclass"]])
        iptde = np.array([bool(str(pc) == "TDE") for pc in metrics_flex["photoclass"]])
        iptde_m = np.array(
            [bool(str(pc) == "TDE?") for pc in metrics_flex["photoclass"]]
        )
        ipsn = np.array(
            [bool(str(pc)[0:2] == "SN") for pc in metrics_flex["photoclass"]]
        )
        ipsn_m = np.array(
            [bool(str(pc)[0:3] == "SN?") for pc in metrics_flex["photoclass"]]
        )

        # decrease rank for fainter reference magnitude
        # removed this in Dec 2020 to decrease host bias
        # rank += np.clip(metrics['ref_mag'].values, 17,25)

        # photmetric TDE get higer rank
        rank[iptde] -= 5
        # maybe photmetric TDE gets somewhat higer rank
        rank[iptde_m] -= 4

        # poor fit gets a lower rank
        rank[ipbad] += 1

        # photmetric AGN get lower rank
        rank[ipagn] += 5

        # CV get lower rankp
        rank[ipcv] += 5

        # photometric SN get lower rank
        rank[ipsn] += 5

        # photometric potential SN get somewhat lower rank
        rank[ipsn_m] += 3

        # increase rank value by number of days we didnt see any detections
        rank += np.clip(metrics["age"], 3, 1000) / 10

        # small preference for source with more detections
        rank -= np.clip(metrics["ndetections"], 1, 10) / 10

        # strong anti preference for sources with many negative subtractions
        rank += (
            metrics["ndetections_negative"]
            / np.clip(metrics["ndetections"], 1, 1e99)
            * 10
        )

        # no ranking based on offset for now, because symatics can be large
        # rank += np.clip(metrics['offset_sig'], 2., 20)
        # rank += np.clip(metrics['offset_sig_ps1'], 2., 20)

        # brighter source somewhat higher rank
        rank += np.clip(metrics["peak_mag"] - 19, 0, 3)

        # linear increase with magnitude flare-host mag difference
        rank += np.clip(metrics["latest_diff_mag"], -2, 5) / 2

        # small extra penalty for upper limits in last detection
        # isignon = (metrics['lastest_obs_is_detection']==0) & (metrics['latest_maglim']>20)
        # self.logger.info ('# sources with last obs is nondetection {}'.format(sum(isignon)))
        # rank[isignon] +=0.5

        # number one gets rank=1
        rank -= np.nanmin(rank) - 1

        return rank

    # This print function needs to also show some info about current classification and current followup status
    def print_ranking(self, filename, metrics, metrics_flex):
        """ """

        # header
        ss = "#name           rank   mag     diff     age   neg/pos  off_stat_ztf off_stat_ps1  mean_color  color_slope     rise       fade   photo_class  official_class    more info"

        for i in np.argsort(metrics["ranking"]):

            ss += "\n{0} {1:6.1f}   {2:5.2f}   {3:5.2f}    {4:5.1f}   {5:5.2f}     {6:7.1f}     {7:7.1f}       {8:5.2f}      {9:7.4f}     {10:9.1f}  {11:9.1f}   {12:14}    {13:7}    {14}".format(
                str(metrics_flex.iloc[i]["name"]),
                float(metrics["ranking"].iloc[i]),
                float(metrics.iloc[i]["latest_mag"]),
                float(metrics.iloc[i]["latest_diff_mag"]),
                float(metrics.iloc[i]["age"]),
                float(
                    metrics.iloc[i]["ndetections_negative"]
                    / np.clip(metrics.iloc[i]["ndetections"], 1, 1e99)
                ),
                float(metrics.iloc[i]["offset_sig"]),
                float(metrics.iloc[i]["offset_sig_ps1"]),
                float(metrics_flex.iloc[i]["mean_color"]),
                float(metrics_flex.iloc[i]["color_slope"]),
                np.clip(float(metrics_flex.iloc[i]["sigma_rise"]), 0, 999),
                np.clip(float(metrics_flex.iloc[i]["sigma_fade"]), 0, 999),
                str(metrics_flex.iloc[i]["photoclass"]),
                str(metrics.iloc[i]["classification"]),
                str(metrics.iloc[i]["extra_info"]),
            )

        self.put(filename, bytes(ss + "\n", "utf-8"))

    def make_sumplots(self, metrics, metrics_flex):
        """
        make summary plot
        currently hardcoded to work for ZTFBH Nuclear
        """

        from numpy import log10, sqrt, log

        inotbogus = np.array(
            [
                bool(
                    (rc != "star")
                    & (rc != "bogus")
                    & (rc != "varstar")
                    & (rc != "star?")
                    & (rc != "stellar")
                )
                for rc in metrics["classification"].values
            ]
        )

        results = metrics_flex[inotbogus].copy()
        if len(results) == 0:
            return

        # some cuts on the Fritz/Marshal/TNS/catalog labels (from classifyme)
        iquestion = np.array(
            [
                ("?" in c) and (c != "TDE?")
                for c in results["classification"].astype(str)
            ]
        )
        iagn = (results["classification"] == "AGN") | (
            results["classification"] == "QSO"
        )
        isnIa = np.array(
            [c[0:5] == "SN Ia" for c in results["classification"].astype(str)]
        ) & (iquestion == False)
        isn_all = np.array(
            [c[0:2] == "SN" for c in results["classification"].astype(str)]
        ) & (iquestion == False)
        isn = isn_all & (isnIa == False)
        icv = np.array([c[0:2] == "CV" for c in results["classification"].astype(str)])
        inone = (
            np.array([c == "None" for c in results["classification"].astype(str)])
            | iquestion
        )

        self.logger.info("\n\n----")
        for ii, ttype in zip(
            *([iagn, isnIa, isn, icv, inone], ("AGN", "Ia", "SN other", "CV", "none"))
        ):
            self.logger.info(f"{ttype}: {sum(ii)}")
        self.logger.info("----\n\n")

        # add this back in later:
        itde = np.array(
            [c[0:3] == "TDE" for c in results["classification"].astype(str)]
        )

        # rename for GOT
        # SvV to do: make a version of the summary plot with all the TDEs
        # GoT_dict = self.read_file('/mampel/misc/GOTnames.json')[1].json()

        # for i, source in results.iterrows():
        #   if source['name'] in GoT_dict.keys():
        #       source['name'] = GoT_dict[source['name']]

        # add marshal class to name (for plotting)
        results["name"] = results.apply(
            lambda r: r["name"] + "(" + r["classification"] + ")"
            if (not "(" in r["name"])
            and (r["classification"] != "TDE")
            and (r["classification"] != "None")
            else r["name"],
            axis=1,
        )

        plot_tuple = (
            [itde, iagn, isnIa, isn, isn_all, inone],
            ("TDE", "AGN", "SN Ia", "SN other", "SN all", "unknown"),
            (1, 0.3, 0.5, 0.8, 0.8, 0.5, None, 0.9),
            ("d", ".", "s", "o", None, "x"),
        )
        mew = 0.6
        lw = 0.6
        ms = 2

        # ---

        # ----
        # color maps obsession
        # cmap = mpl.cm.get_cmap('magma')

        cmap = mpl.cm.get_cmap("viridis")
        manual_colors = [cmap(0.1)] + [
            cmap(0.65),
            cmap(0.65),
            cmap(0.65),
            cmap(0.35),
            cmap(0.92),
        ]
        labels = ["unknown", "SN other", "SN all", "SN Ia", "TDE", "AGN"]
        color_dict = dict(zip(labels, manual_colors))

        if self.plotHisto:
            plt.clf()
            ipl = results["mag_peak"] > 0
            plt.title("all sources; N=" + str(sum(ipl)))
            plt.hist(results[ipl]["mag_peak"], range=[17, 22], bins=20)
            plt.xlabel("peak mag")

            plot_fname = self.plot_dir + "/all_maghisto.pdf"
            self.plt2box(plt, plot_fname)

        # ------
        # rise / fade
        finfo = "{0:30} {1:10} {2:10}\n".format("name", "rise_time", "fade_rime")

        ipl = (
            (results["n_rise"] > 2)
            & (results["n_fade"] > 2)
            & (results["e_sigma_fade"] > 0)
            & (results["sigma_rise"] / results["e_sigma_rise"] > 1.5)
            & (results["sigma_rise"] > 1.1)
            & (results["sigma_fade"] / results["e_sigma_fade"] > 1.5)
        )

        if self.plotHisto:
            plt.clf()
            plt.title("rise/fade selection; N=" + str(sum(ipl)))
            plt.hist(results[ipl]["mag_peak"], range=[17, 22], bins=20)
            plt.xlabel("peak mag")

            plot_fname = self.plot_dir + "/rise_fade_maghisto.pdf"
            self.plt2box(plt, plot_fname)

        reference_agn = pd.read_csv(
            io.BytesIO(self.read_file("/mampel/misc/AGN.csv").content)
        )
        reference_sne = pd.read_csv(
            io.BytesIO(self.read_file("/mampel/misc/SN_all.csv").content)
        )
        reference_tde = pd.read_csv(
            io.BytesIO(self.read_file("/mampel/misc/TDE.csv").content)
        )
        plt.clf()
        for idx, labl, alph, fmt in zip(*plot_tuple):

            self.logger.info(f"{labl}, {sum(ipl&idx)}")
            x, y = log10(results[ipl & idx]["sigma_rise"].astype(np.float64)) + log10(
                sqrt(2)
            ), log10(results[ipl & idx]["sigma_fade"].astype(np.float64))

            xerr = (
                1
                / log(10)
                * results[ipl & idx]["e_sigma_rise"].values
                / results[ipl & idx]["sigma_rise"].values
            )
            yerr = (
                1
                / log(10)
                * results[ipl & idx]["e_sigma_fade"].values
                / results[ipl & idx]["sigma_fade"].values
            )
            x, y = x.values, y.values

            if fmt:
                line = plt.errorbar(
                    x,
                    y,
                    xerr=xerr,
                    yerr=yerr,
                    color=color_dict[labl],
                    fmt=fmt,
                    alpha=alph,
                    label=labl,
                    lw=lw,
                    mew=mew,
                    ms=ms,
                    zorder=3 + 1 / (sum(idx) + 1),
                )

            if labl == "TDE":
                for i, nm in enumerate(results[ipl & idx]["name"]):
                    # if nm=='NedStark': nm='AT2018zr'
                    finfo += "{0:26} {1:10.3f}     {2:10.3f}\n".format(nm, x[i], y[i])
                    plt.text(x[i], y[i], nm, fontsize=4, zorder=5)
            elif labl == "unknown":
                for i, nm in enumerate(results[ipl & idx]["name"]):
                    if x[i] > 0 or y[i] > 1.0:
                        finfo += "{0:26} {1:10.3f}     {2:10.3f}\n".format(
                            nm, x[i], y[i]
                        )
                        plt.text(x[i], y[i], nm, fontsize=3, zorder=6)
            elif labl != "AGN":
                for i, nm in enumerate(results[ipl & idx]["name"]):
                    if (x[i] > 0 or y[i] > 1.0) or ("ZTF" in nm is False):
                        finfo += "{0:26} {1:10.3f}     {2:10.3f}\n".format(
                            nm, x[i], y[i]
                        )
                        plt.text(x[i], y[i], nm, fontsize=2, zorder=5)
        ax = plt.gca()
        corner.hist2d(
            log10(reference_agn["sigma_rise"].values) + log10(sqrt(2)),
            log10(reference_agn["sigma_fade"].values),
            bins=[35, 35],
            range=[(0, log10(500)), (0, log10(500))],
            smooth=True,
            levels=([0.66, 0.66, 0.66]),
            ax=ax,
            color=color_dict["AGN"],
            plot_datapoints=False,
            plot_density=False,
            plot_contours=True,
            no_fill_contours=True,
            fill_contours=False,
            contour_kwargs={"zorder": 0},
        )

        corner.hist2d(
            log10(reference_sne["sigma_rise"].values) + log10(sqrt(2)),
            log10(reference_sne["sigma_fade"].values),
            bins=[35, 35],
            range=[(0, log10(500)), (0, log10(500))],
            smooth=True,
            levels=([0.66, 0.66, 0.66]),
            ax=ax,
            color=color_dict["SN all"],
            plot_datapoints=False,
            plot_density=False,
            plot_contours=True,
            no_fill_contours=True,
            fill_contours=False,
            contour_kwargs={"zorder": 0},
        )

        plt.legend(loc=4)
        plt.xlabel(r"rise e-folding time (log day)")
        plt.ylabel(r"fade e-folding time (log day)")

        plt.xlim(0.7, 2.5)
        plt.ylim(1.0, 2.5)

        plot_fname = self.plot_dir + "/rise_fade.pdf"
        self.plt2box(plt, plot_fname)

        # also plot the known TDEs
        plot_fname = plot_fname.replace(".pdf", "_wtdes.pdf")
        line = plt.errorbar(
            log10(reference_tde["sigma_rise"].values) + log10(sqrt(2)),
            log10(reference_tde["sigma_fade"].values),
            color=color_dict["TDE"],
            fmt="d",
            alpha=0.7,
            lw=lw,
            mew=mew,
            ms=ms,
            zorder=1,
        )
        self.plt2box(plt, plot_fname)

        # ------
        # color / cooling
        finfo += "\n\n {0:30} {1:10} {2:10}\n".format(
            "name", "mean_color", "color_change"
        )
        ipl = (
            (results["mean_color"] != 0)
            & (np.abs(results["e_mean_color"]) < 0.35)
            & (results["n_fade"] > 3)
            & (np.abs(results["e_color_slope"]) < 0.03)
        )

        if self.plotHisto:
            plt.clf()
            plt.title("color/cooling selection; N=" + str(sum(ipl)))
            plt.hist(results[ipl]["mag_peak"], range=[17, 22], bins=20)
            plt.xlabel("peak mag")

            plot_fname = self.plot_dir + "/color_change_maghisto.pdf"
            self.plt2box(plt, plot_fname)

        plt.clf()
        for idx, labl, alph, fmt in zip(*plot_tuple):

            self.logger.info(f"{labl}, {sum(ipl&idx)}")
            x, y = results[ipl & idx]["mean_color"], results[ipl & idx]["color_slope"]
            x, y = x.values, y.values  # to np array

            # clip to keep very blue sources in range
            x = np.clip(x, -0.6, 0.9)

            if fmt:
                line = plt.errorbar(
                    x,
                    y,
                    xerr=results[ipl & idx]["e_mean_color"],
                    yerr=results[ipl & idx]["e_color_slope"],
                    color=color_dict[labl],
                    fmt=fmt,
                    alpha=alph,
                    lw=lw,
                    mew=mew,
                    ms=ms,
                    label=labl,
                    zorder=3 + 1 / (sum(idx) + 1),
                )

            if labl == "TDE":
                for i, nm in enumerate(results[ipl & idx]["name"]):
                    finfo += "{0:26} {1:10.3f}     {2:10.3f}\n".format(nm, x[i], y[i])
                    plt.text(x[i] - 0.01, y[i], nm, fontsize=1.5, zorder=5)
            elif (labl != "AGN") and (labl[0:2] != "SN"):
                for i, nm in enumerate(results[ipl & idx]["name"]):
                    if (
                        (y[i] < 0.02)
                        and not ("AGN?" in nm)
                        and not ("CV?" in nm)
                        and not ("varstar?" in nm)
                        or (x[i] <= 0.4)
                    ):

                        if "?" in nm:
                            fs = 1
                        else:
                            fs = 1.5
                            finfo += "{0:26} {1:10.3f}     {2:10.3f}\n".format(
                                nm, x[i], y[i]
                            )
                            plt.text(x[i], y[i], nm, fontsize=fs, zorder=5)
                        self.logger.info(f"{nm}, {not('AGN?' in nm)}")

        ax = plt.gca()
        corner.hist2d(
            reference_sne["mean_color"].values,
            reference_sne["color_slope"].values,
            bins=[10, 20],
            range=[(-0.4, 0.8), (-0.03, +0.045)],
            smooth=True,
            levels=([0.66, 0.66, 0.66]),
            ax=ax,
            color=color_dict["SN all"],
            plot_datapoints=False,
            plot_density=False,
            plot_contours=True,
            no_fill_contours=True,
            fill_contours=False,
            contour_kwargs={"zorder": 0},
        )
        plt.legend()
        plt.xlabel("mean g-r color")
        plt.ylabel("rate of color change (1/day)")
        plt.xlim(-0.6, 0.8)
        plt.ylim(-0.020, 0.04)
        # plt.draw()
        # plt.pause(0.05)

        # option to add a "TDE box"
        # plt.plot([-1,-0.00],[0.015,0.015], '--k',alpha=0.6, lw=0.5)
        # plt.plot([-0.1,-0.00],[-1,0.015], '--k',alpha=0.6, lw=0.5)
        # txt = plt.text(-0.3, -0.01, 'TDE box',  fontsize='large', verticalalignment='center', horizontalalignment='center')
        # txt.set_bbox(dict(facecolor='white', alpha=0.8, edgecolor='w'))

        plt.title("late-time (post-peak) selection")
        plot_fname = self.plot_dir + "/color_change.pdf"
        self.plt2box(plt, plot_fname)

        # also plot the known TDEs
        plot_fname = plot_fname.replace(".pdf", "_wtdes.pdf")
        line = plt.errorbar(
            reference_tde["mean_color"].values,
            reference_tde["color_slope"].values,
            color=color_dict["TDE"],
            fmt="d",
            alpha=0.7,
            lw=lw,
            mew=mew,
            ms=ms,
            zorder=1,
        )
        self.plt2box(plt, plot_fname)

        # ------
        # rise / color
        # print ('\n\n rise/color')
        finfo += "\n\n {0:30} {1:10} {2:10}\n".format("name", "mean_color", "rise_time")
        ipl = (
            (results["mean_color"] != 0)
            & (results["n_rise"] > 2)
            & (results["sigma_rise"] / results["e_sigma_rise"] > 1.5)
            & (abs(results["e_mean_color"]) < 0.35)
        )

        if self.plotHisto:
            plt.clf()
            plt.title("rise-time + color selection selection; N=" + str(sum(ipl)))
            plt.hist(results[ipl]["mag_peak"], range=[17, 22], bins=10)
            plt.xlabel("peak mag")

            plot_fname = self.plot_dir + "/rise_color_maghisto.pdf"
            self.plt2box(plt, plot_fname)

        plt.clf()
        for idx, labl, alph, fmt in zip(*plot_tuple):

            self.logger.info(f"{labl}, {sum(ipl&idx)}")

            x, y = results[ipl & idx]["mean_color"], log10(
                results[ipl & idx]["sigma_rise"]
            ) + log10(sqrt(2))
            yerr = (
                1
                / log(10)
                * results[ipl & idx]["e_sigma_rise"]
                / results[ipl & idx]["sigma_rise"]
            )
            x, y = x.values, y.values

            # clip to keep very blue sources in the plot range
            x = np.clip(x, -0.6, 0.9)

            if fmt:
                line = plt.errorbar(
                    x,
                    y,
                    xerr=results[ipl & idx]["e_mean_color"],
                    yerr=yerr,
                    color=color_dict[labl],
                    fmt=fmt,
                    alpha=alph,
                    label=labl,
                    lw=lw,
                    mew=mew,
                    ms=ms,
                    zorder=3 + 1 / (sum(idx) + 1),
                )

            if labl == "TDE":
                for i, nm in enumerate(results[ipl & idx]["name"]):
                    finfo += "{0:26} {1:10.3f}     {2:10.3f}\n".format(nm, x[i], y[i])
                    plt.text(x[i], y[i], nm, fontsize=2, zorder=5)
            elif labl != "AGN":
                for i, nm in enumerate(results[ipl & idx]["name"]):
                    if ((x[i] < 0.3 and y[i] > 0) and not ("SN" in nm)) or (
                        x[i] < -0.1 or y[i] > 0.5
                    ):

                        if ("?" in nm) or ("Ia" in nm):
                            fs = 1
                        else:
                            fs = 1.6
                        finfo += "{0:26} {1:10.3f}     {2:10.3f}\n".format(
                            nm, x[i], y[i]
                        )
                        plt.text(x[i], y[i], nm, fontsize=fs, zorder=6)

        ax = plt.gca()
        corner.hist2d(
            reference_agn["mean_color"].values,
            log10(reference_agn["sigma_rise"].values) + log10(sqrt(2)),
            bins=[30, 30],
            range=[(-0.6, 0.8), (0, log10(300))],
            smooth=True,
            levels=([0.66, 0.66, 0.66]),
            ax=ax,
            color=color_dict["AGN"],
            plot_datapoints=False,
            plot_density=False,
            plot_contours=True,
            no_fill_contours=True,
            fill_contours=False,
            contour_kwargs={"zorder": 0},
        )  # 'lw':1 keyword not used by corner
        corner.hist2d(
            reference_sne["mean_color"].values,
            log10(reference_sne["sigma_rise"].values) + log10(sqrt(2)),
            bins=[30, 30],
            range=[(-0.6, 0.8), (0, log10(300))],
            smooth=True,
            levels=([0.66, 0.66, 0.66]),
            ax=ax,
            color=color_dict["SN all"],
            plot_datapoints=False,
            plot_density=False,
            plot_contours=True,
            no_fill_contours=True,
            fill_contours=False,
            contour_kwargs={"zorder": 0},
        )

        plt.title("real-time (pre-peak) selection")
        plt.legend(loc=1)
        plt.xlabel("mean g-r color")
        plt.ylabel(r"rise e-folding time (log$_{10}$ day)")
        plt.xlim(-0.6, 0.8)
        plt.ylim(log10(4), log10(300))

        plot_fname = self.plot_dir + "/rise_color.pdf"
        self.plt2box(plt, plot_fname)

        # also plot the known TDEs
        plot_fname = plot_fname.replace(".pdf", "_wtdes.pdf")
        line = plt.errorbar(
            reference_tde["mean_color"].values,
            log10(reference_tde["sigma_rise"].values) + log10(sqrt(2)),
            color=color_dict["TDE"],
            fmt="d",
            alpha=0.7,
            lw=lw,
            mew=mew,
            ms=ms,
            zorder=1,
        )
        self.plt2box(plt, plot_fname)

        # as a bonus: write a file with the names of TDE candidates (based on some boxing)
        # for each of the plots (requested by Suvi)
        self.put(self.plot_dir + "/boxed.txt", bytes(finfo, "utf-8"))

    # pyplot to dropbox function
    def plt2box(self, plt, fname):
        """
        write PDF of plot in pyplot to dropbox
        (could be more neat to work with ax objects everywhere, but hey)
        >> plt2box(plt, plot_name)
        """
        buf = io.BytesIO()
        ax = plt.gca()
        ax.figure.savefig(buf, format="pdf")
        buf.seek(0)
        self.put(fname, buf.read())

    # some last helper functions
    def datetomjd(self, d):
        d0 = datetime.datetime(1858, 11, 17, 0, 0, 0)
        dt = d - d0
        # dt is a timedelta object.
        return dt.days + (dt.seconds + dt.microseconds / 1e6) / 86400.0

    def mjdtodate(self, mjd):
        if not np.isfinite(mjd):
            return None
        jd = mjd + 2400000.5
        unixtime = (jd - 2440587.5) * 86400.0  # in seconds
        return datetime.datetime.utcfromtimestamp(unixtime)

    def format_date(self, transient_date):
        return astropy.time.Time(
            transient_date[6:10]
            + "-"
            + transient_date[3:5]
            + "-"
            + transient_date[0:2]
            + transient_date[10:]
        ).jd



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : ampel/nuclear/t2/T2FlexFit.py
# License           : BSD-3-Clause
# Author            : mitchell@nyu.edu
# Date              : 28.04.2020
# Last Modified Date: 08.06.2020
# Last Modified By  : mitchell@nyu.edu

import sys
import numpy as np
import logging

logging.basicConfig()

from ampel.nuclear.flexfit import flex_metrics
from ampel.abstract.AbsLightCurveT2Unit import AbsLightCurveT2Unit
from ampel.view import LightCurve
from ampel.types import UBson
from ampel.struct.UnitResult import UnitResult
from ampel.ztf.util.ZTFIdMapper import to_ztf_id


class T2FlexFit(AbsLightCurveT2Unit):
    """
    Least squares fitting on (ZTF) data with observations in two bands
    We fit for the peak, using a Gaussian rise and an exponential decay
    We measure the mean color pre-peak, and allow for a color-change post-peak

    Goal is photometric typing of alerts into:

     - "SN" or "Ia" : based on rise time
     - "AGN"        : slow rise or fading
     - "CV"         : fast
     - "TDE"        : not the above and constant color or blue
    """

    oldest_upper_limits: int
    max_post_peak: int

    def process(self, light_curve) -> UBson | UnitResult:
        """
        Parameters
        ----------
        light_curve: `ampel.view.LightCurve` instance.
        See original docstring & photopoint docstring.
        Note: arguments of '.get_values()' method are strings and are in the following table: https://zwickytransientfacility.github.io/ztf-avro-alert/schema.html

        run_config: 'dict'
        'local_data': plotting folder
        'name': ZTF name of object,

        Returns
        -------
        Dictionary of three dicts: fit parameters, lightcurve data, and plotting info:
        {'fit_params':
        ['dtime_peak', 'band', 'mag_peak', 'sigma_rise', 'sigma_fade', 'mean_color', 'color_slope', 'e_dtime_peak',
        'e_mag_peak', 'e_mean_color', 'e_color_slope', 'e_sigma_rise', 'mad', 'rms', 'chi2', 'n_rise', 'n_fade',
        'classification', 'name', 'mjd_peak', 'photoclass', 'blabber'],

        'lc_data':
        ['dtime', 'flux', 'flux_err', 'fid', 'isdetect', 'jd', 'jd0'],

        'plot_info':
        ['lsq_out', 'dt_fix', 'dtime_out', 'flux_out', , 'pid', 'programid', 'finfo']
        }

        """
        lc = get_raw_lc(light_curve)
        name = to_ztf_id(light_curve.get_photopoints()[0]["stock"][0])
        meta_data = {
            "name": name,
            "oldest_upper_limits": self.oldest_upper_limits,
            "max_post_peak": self.max_post_peak,
            "logger": self.logger,
        }
        result = flex_metrics.compute(meta_data, lc)
        return jsonify(result)


def jsonify(obj):
    """
    Recursively cast to JSON native types
    """
    if hasattr(obj, "tolist"):
        return obj.tolist()
    elif hasattr(obj, "keys"):
        return {jsonify(k): jsonify(obj[k]) for k in obj.keys()}
    elif hasattr(obj, "__len__") and not isinstance(obj, str):
        return [jsonify(v) for v in obj]
    else:
        return obj


def _lc_to_dict(lc):
    # Turns Ampel Lightcurve object into a dict
    keys = (
        "jd",
        "fid",
        "pid",
        "diffmaglim",
        "pdiffimfilename",
        "programpi",
        "programid",
        "candid",
        "isdiffpos",
        "tblid",
        "nid",
        "rcid",
        "field",
        "xpos",
        "ypos",
        "ra",
        "dec",
        "magpsf",
        "sigmapsf",
        "chipsf",
        "magap",
        "sigmagap",
        "distnr",
        "magnr",
        "sigmagnr",
        "chinr",
        "sharpnr",
        "sky",
        "magdiff",
        "fwhm",
        "classtar",
        "mindtoedge",
        "magfromlim",
        "seeratio",
        "aimage",
        "bimage",
        "aimagerat",
        "bimagerat",
        "elong",
        "nneg",
        "nbad",
        "rb",
        "rbversion",
        "drb",
        "drbversion",
        "ssdistnr",
        "ssmagnr",
        "ssnamenr",
        "sumrat",
        "magapbig",
        "sigmagapbig",
        "ranr",
        "decnr",
        "ndehist",
        "ncovhist",
        "jdstarhist",
        "jdendhist",
        "scorr",
        "tooflag",
        "objectidps1",
        "sgmag1",
        "srmag1",
        "simag1",
        "szmag1",
        "sgscore1",
        "distpsnr1",
        "objectidps2",
        "sgmag2",
        "srmag2",
        "simag2",
        "szmag2",
        "sgscore2",
        "distpsnr2",
        "objectidps3",
        "sgmag3",
        "srmag3",
        "simag3",
        "szmag3",
        "sgscore3",
        "distpsnr3",
        "nmtchps",
        "rfid",
        "jdstartref",
        "jdendref",
        "nframesref",
        "dsnrms",
        "ssnrms",
        "dsdiff",
        "magzpsci",
        "magzpsciunc",
        "magzpscirms",
        "nmatches",
        "clrcoeff",
        "clrcounc",
        "zpclrcov",
        "zpmed",
        "clrmed",
        "clrrms",
        "neargaia",
        "neargaiabright",
        "maggaia",
        "maggaiabright",
        "exptime",
    )

    vals = []

    for key in keys:
        vals.append(np.array(lc.get_values(key)))

    return dict(zip(keys, vals))


def _append_limits(lc, lcurve):
    # lc is dict, lcurve is LightCurve instance
    # If there's a more efficient way to do this with Ampel let us know!
    if lcurve.get_values("jd", of_upper_limits=True):
        ul_jd = lcurve.get_values("jd", of_upper_limits=True)
        ul_fid = lcurve.get_values("fid", of_upper_limits=True)
        ul_diffmaglim = lcurve.get_values("diffmaglim", of_upper_limits=True)
        ul_isdiffpos = np.full(len(ul_jd), "t")

        lc["jd"] = np.append(lc["jd"], ul_jd)
        lc["fid"] = np.append(lc["fid"], ul_fid)
        lc["diffmaglim"] = np.append(lc["diffmaglim"], ul_diffmaglim)
        lc["isdiffpos"] = np.append(lc["isdiffpos"], ul_isdiffpos)

        for key in lc:
            if key not in ["jd", "fid", "diffmaglim", "isdiffpos"]:
                lc[key] = np.append(lc[key], np.full(len(ul_jd), 999))

    return lc


def get_raw_lc(light_curve):
    lc = _lc_to_dict(light_curve)
    lc = _append_limits(lc, light_curve)
    return lc



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : ampel/nuclear/t3/T3MetricsPlots.py
# License           : BSD-3-Clause
# Author            : mitchell@nyu.edu
# Date              : 08.06.2020
# Last Modified Date: 29.11.2022
# Last Modified By  : simeon.reusch@desy.de


import os, sys, datetime, io, json, warnings
from collections.abc import Generator
from typing import Any, Optional, Union

from pydantic import BaseModel
import matplotlib as mpl  # type: ignore
import corner  # type: ignore
import astropy  # type: ignore
from numpy import log10, sqrt, log
import numpy as np
import pandas as pd  # type: ignore
from matplotlib import pyplot as plt

from ampel.types import UBson, T3Send
from ampel.view.TransientView import TransientView
from ampel.view.T3Store import T3Store
from ampel.struct.UnitResult import UnitResult
from ampel.abstract.AbsPhotoT3Unit import AbsPhotoT3Unit
from ampel.nuclear.t3.dropboxIO import DropboxUnit
import ampel.nuclear.t3.classifyme as classifyme

warnings.filterwarnings("ignore", category=RuntimeWarning)

flex_keys = [
    "dtime_peak",
    "band",
    "mag_peak",
    "sigma_rise",
    "sigma_fade",
    "mean_color",
    "color_slope",
    "e_dtime_peak",
    "e_mag_peak",
    "e_mean_color",
    "e_color_slope",
    "e_sigma_rise",
    "e_sigma_fade",
    "mad",
    "rms",
    "chi2",
    "n_rise",
    "n_fade",
    "classification",
    "name",
    "mjd_peak",
    "photoclass",
    "blabber",
]
simple_keys = [
    "name",
    "classification",
    "ra",
    "dec",
    "ndetections",
    "ndetections_negative",
    "peak_mag",
    "latest_mag",
    "peak_jd",
    "latest_jd",
    "age",
    "ndetections_g",
    "offset_rms_g",
    "offset_unc_g",
    "offset_med_g",
    "offset_weighted_g",
    "offset_sig_g",
    "ndetections_r",
    "offset_rms_r",
    "offset_unc_r",
    "offset_med_r",
    "offset_weighted_r",
    "offset_sig_r",
    "offset_sig",
    "lastest_obs_is_detection",
    "latest_maglim",
    "ref_mag_r",
    "ref_mag_g",
    "peak_mag_g",
    "latest_mag_g",
    "peak_diff_mag_g",
    "latest_diff_mag_g",
    "peak_mag_r",
    "latest_mag_r",
    "peak_diff_mag_r",
    "latest_diff_mag_r",
    "ref_mag",
    "latest_diff_mag",
    "peak_diff_mag",
]


class T3SummaryPlots(DropboxUnit, AbsPhotoT3Unit):

    plotHisto: bool = False  # create histrograms for of the magnitude of the transients in the different summary plots
    force_date: Optional[str] = None  # force the save date for testing

    def post_init(self):
        super().post_init()
        self.save_location = self.base_location + "/sum_plots"
        if "sum_plots" not in self.get_files(self.base_location):
            self.create_folder(self.save_location)

    def plt2box(self, plt, fname):
        """
        write PDF of plot in pyplot to dropbox
        (could be more neat to work with ax objects everywhere, but hey)
        >> plt2box(plt, plot_name)
        """
        buf = io.BytesIO()
        ax = plt.gca()
        ax.figure.savefig(buf, format="pdf")
        buf.seek(0)
        self.put(fname, buf.read())

    def process(
        self, gen: Generator[TransientView, T3Send, None], t3s: Optional[T3Store] = None
    ) -> Union[UBson, UnitResult]:
        """ """
        # DUMMY FUNCTION FOR NOW
        return None

    def add(self, transients):
        """
        print some detection stats
        and make summary plost
        currently hardcoded to work for ZTFBH Nuclear
        date_str='yyyy-mm-dd', optional input to selected what date to print
        days_back=60, days since the last detection  to be included in the plot
        """

        drop_dir = self.save_location + "/" + str(self.night.datetime.year)
        if str(self.night.datetime.year) not in self.get_files(self.save_location):
            self.create_folder(drop_dir)

        plot_dir = drop_dir + "/" + self.night.datetime.strftime("%m-%d")
        if self.night.datetime.strftime("%m-%d") not in self.get_files(drop_dir):
            self.create_folder(plot_dir)

        metrics, metrics_flex = self.collect_metrics(transients)

        inotbogus = np.array(
            [
                bool(
                    (rc != "star")
                    & (rc != "bogus")
                    & (rc != "varstar")
                    & (rc != "star?")
                    & (rc != "stellar")
                )
                for rc in metrics["classification"].values
            ]
        )

        mjdmay = self.datetomjd(datetime.datetime(2018, 5, 1))
        mjdnov = self.datetomjd(datetime.datetime(2018, 10, 1))
        jdnow = self.datetomjd(self.night.datetime) + 2400000.5
        mjdmax = np.nanmax(metrics_flex["mjd_peak"])

        self.logger.info(
            "date of max mjd_peak from fit: " + str(self.mjdtodate(mjdmax))
        )

        number_of_days = mjdmax - mjdnov
        ipostnov = (
            metrics_flex["mjd_peak"].values > mjdnov
        )  # used for printing detection stat
        irecent = metrics["latest_jd"].values > -999  # select all, deprecated daysback

        results = metrics_flex.iloc[ipostnov & inotbogus].copy()
        self.logger.info("# of days since 2018-Nov-1: " + str(number_of_days))

        mag_lim = np.arange(20, 18, -0.5)
        iokrise = results["n_rise"] > 2

        self.logger.info("\n>2 pre-peak detections & not AGN:")

        for magl in mag_lim:
            ilim = results["mag_peak"] < magl
            rr = results[iokrise & ilim].copy()
            inotagn = (rr["photoclass"] != "AGN?") & (rr["classification"] != "AGN")
            self.logger.info(
                "{0:0.1f} {1:0.1f}/week".format(magl, sum(inotagn) / number_of_days * 7)
            )

        self.logger.info("\n>2 pre-peak detections & not AGN & not photo SNe:")

        for magl in mag_lim:
            ilim = results["mag_peak"] < magl
            rr = results[iokrise & ilim].copy()
            inotagn = (rr["photoclass"] != "AGN?") & (rr["classification"] != "AGN")
            inotsn = rr["photoclass"] == "not Ia?"
            self.logger.info(
                "{0:0.1f} {1:0.1f}/week".format(
                    magl, sum(inotsn & inotagn) / number_of_days * 7
                )
            )

        self.logger.info("\n\n>3 post-peak detections & not AGN:")
        iokfade = results["n_fade"] > 3
        for magl in mag_lim:
            ilim = results["mag_peak"] < magl
            rr = results[iokfade & ilim].copy()
            inotagn = (rr["photoclass"] != "AGN?") & (rr["classification"] != "AGN")
            self.logger.info(
                "{0:0.1f} {1:0.1f}/week".format(magl, sum(inotagn) / number_of_days * 7)
            )

        self.logger.info(
            "\n\n>3 post-peak detections & TDE candidate (blue or not cooling):"
        )
        iokfade = results["n_fade"] > 3
        for magl in mag_lim:
            ilim = results["mag_peak"] < magl
            rr = results[iokfade & ilim].copy()
            itde = [bool("TDE" in pc) for pc in rr["photoclass"] if "bad" not in pc]
            self.logger.info(
                "{0:0.1f} {1:0.1f}/week".format(magl, sum(itde) / number_of_days * 7)
            )

        self.logger.info("\n>10 post-peak detections & TDE candidate & not red")
        iokfade = results["n_fade"] > 3
        for magl in mag_lim:
            ilim = results["mag_peak"] < magl
            rr = results[iokfade & ilim].copy()
            itde = [
                "TDE" in pc
                for pc in rr["photoclass"]
                if "red" not in pc and "bad" not in pc
            ]
            self.logger.info(
                "{0:0.1f} {1:0.1f}/week".format(magl, sum(itde) / number_of_days * 7)
            )

        results = metrics_flex[irecent & inotbogus].copy()
        if len(results) == 0:
            return

        iquestion = np.array(
            [
                ("?" in c) and (c != "TDE?")
                for c in results["classification"].astype(str)
            ]
        )
        iagn = (results["classification"] == "AGN") | (
            results["classification"] == "QSO"
        )
        isnIa = np.array(
            [c[0:5] == "SN Ia" for c in results["classification"].astype(str)]
        ) & (iquestion == False)
        isn_all = np.array(
            [c[0:2] == "SN" for c in results["classification"].astype(str)]
        ) & (iquestion == False)
        isn = isn_all & (isnIa == False)
        icv = np.array([c[0:2] == "CV" for c in results["classification"].astype(str)])
        inone = (
            np.array([c == "None" for c in results["classification"].astype(str)])
            | iquestion
        )

        self.logger.info("\n\n----")
        for ii, ttype in zip(
            *([iagn, isnIa, isn, icv, inone], ("AGN", "Ia", "SN other", "CV", "none"))
        ):
            self.logger.info(f"{ttype}: {sum(ii)}")
        self.logger.info("----\n\n")

        # add this back in later:
        itde = np.array(
            [c[0:3] == "TDE" for c in results["classification"].astype(str)]
        )

        # rename for GOT
        # SvV to do: make a version of the summary plot with all the TDEs
        GoT_dict = self.read_file("/mampel/misc/GOTnames.json")[1].json()

        for i, source in results.iterrows():
            if source["name"] in GoT_dict.keys():
                source["name"] = GoT_dict[source["name"]]

        # results[results['name']=='ZTF18ablllyw'] = 'ZTF18ablllyw (SN?)' # SN with pre-cursor?!
        # results[results['name']=='ZTF18aayatjf'] = 'ZTF18aayatjf (SN?)' # slow fading SN

        # add marshal class to name (for plotting)
        results["name"] = results.apply(
            lambda r: r["name"] + "(" + r["classification"] + ")"
            if (not "(" in r["name"])
            and (r["classification"] != "TDE")
            and (r["classification"] != "None")
            else r["name"],
            axis=1,
        )

        plot_tuple = (
            [itde, iagn, isnIa, isn, isn_all, inone],
            ("TDE", "AGN", "SN Ia", "SN other", "SN all", "unknown"),
            (1, 0.3, 0.5, 0.8, 0.8, 0.5, None, 0.9),
            ("d", ".", "s", "o", None, "x"),
        )
        mew = 0.6
        lw = 0.6
        ms = 2

        # ---

        # ----
        # color maps obsession
        # cmap = mpl.cm.get_cmap('magma')

        cmap = mpl.cm.get_cmap("viridis")
        manual_colors = [cmap(0.1)] + [
            cmap(0.65),
            cmap(0.65),
            cmap(0.65),
            cmap(0.35),
            cmap(0.92),
        ]
        labels = ["unknown", "SN other", "SN all", "SN Ia", "TDE", "AGN"]
        color_dict = dict(zip(labels, manual_colors))

        if self.plotHisto:
            plt.clf()
            ipl = results["mag_peak"] > 0
            plt.title("all sources; N=" + str(sum(ipl)))
            plt.hist(results[ipl]["mag_peak"], range=[17, 22], bins=20)
            plt.xlabel("peak mag")

            plot_fname = plot_dir + "/all_maghisto.pdf"
            self.plt2box(plt, plot_fname)

        # ------
        # rise / fade
        finfo = "{0:30} {1:10} {2:10}\n".format("name", "rise_time", "fade_rime")
        ipl = (
            (results["n_rise"] > 2)
            & (results["n_fade"] > 2)
            & (results["e_sigma_fade"] > 0)
            & (results["sigma_rise"] / results["e_sigma_rise"] > 1.5)
            & (results["sigma_rise"] > 1.1)
            & (results["sigma_fade"] / results["e_sigma_fade"] > 1.5)
        )

        if self.plotHisto:
            plt.clf()
            plt.title("rise/fade selection; N=" + str(sum(ipl)))
            plt.hist(results[ipl]["mag_peak"], range=[17, 22], bins=20)
            plt.xlabel("peak mag")

            plot_fname = plot_dir + "/rise_fade_maghisto.pdf"
            self.plt2box(plt, plot_fname)

        reference_agn = pd.read_csv(
            io.BytesIO(self.read_file("/mampel/misc/AGN.csv")[1].content)
        )
        reference_sne = pd.read_csv(
            io.BytesIO(self.read_file("/mampel/misc/SN_all.csv")[1].content)
        )
        reference_tde = pd.read_csv(
            io.BytesIO(self.read_file("/mampel/misc/TDE.csv")[1].content)
        )
        plt.clf()
        for idx, labl, alph, fmt in zip(*plot_tuple):

            self.logger.info(f"{labl}, {sum(ipl&idx)}")
            x, y = log10(results[ipl & idx]["sigma_rise"].astype(np.float64)) + log10(
                sqrt(2)
            ), log10(results[ipl & idx]["sigma_fade"].astype(np.float64))

            xerr = (
                1
                / log(10)
                * results[ipl & idx]["e_sigma_rise"].values
                / results[ipl & idx]["sigma_rise"].values
            )
            yerr = (
                1
                / log(10)
                * results[ipl & idx]["e_sigma_fade"].values
                / results[ipl & idx]["sigma_fade"].values
            )
            x, y = x.values, y.values

            if fmt:
                line = plt.errorbar(
                    x,
                    y,
                    xerr=xerr,
                    yerr=yerr,
                    color=color_dict[labl],
                    fmt=fmt,
                    alpha=alph,
                    label=labl,
                    lw=lw,
                    mew=mew,
                    ms=ms,
                    zorder=3 + 1 / (sum(idx) + 1),
                )

            if labl == "TDE":
                for i, nm in enumerate(results[ipl & idx]["name"]):
                    # if nm=='NedStark': nm='AT2018zr'
                    finfo += "{0:26} {1:10.3f}     {2:10.3f}\n".format(nm, x[i], y[i])
                    plt.text(x[i], y[i], nm, fontsize=4, zorder=5)
            elif labl == "unknown":
                for i, nm in enumerate(results[ipl & idx]["name"]):
                    if x[i] > 1.2 or y[i] > 1.6:
                        finfo += "{0:26} {1:10.3f}     {2:10.3f}\n".format(
                            nm, x[i], y[i]
                        )
                        plt.text(x[i], y[i], nm, fontsize=3, zorder=6)
            elif labl != "AGN":
                for i, nm in enumerate(results[ipl & idx]["name"]):
                    if (x[i] > 1.3 or y[i] > 1.5) or ("ZTF" in nm is False):
                        finfo += "{0:26} {1:10.3f}     {2:10.3f}\n".format(
                            nm, x[i], y[i]
                        )
                        plt.text(x[i], y[i], nm, fontsize=2, zorder=5)
        ax = plt.gca()
        corner.hist2d(
            log10(reference_agn["sigma_rise"].values) + log10(sqrt(2)),
            log10(reference_agn["sigma_fade"].values),
            bins=[35, 35],
            range=[(0, log10(500)), (0, log10(500))],
            smooth=True,
            levels=([0.66, 0.66, 0.66]),
            ax=ax,
            color=color_dict["AGN"],
            plot_datapoints=False,
            plot_density=False,
            plot_contours=True,
            no_fill_contours=True,
            fill_contours=False,
            contour_kwargs={"zorder": 0},
        )

        corner.hist2d(
            log10(reference_sne["sigma_rise"].values) + log10(sqrt(2)),
            log10(reference_sne["sigma_fade"].values),
            bins=[35, 35],
            range=[(0, log10(500)), (0, log10(500))],
            smooth=True,
            levels=([0.66, 0.66, 0.66]),
            ax=ax,
            color=color_dict["SN all"],
            plot_datapoints=False,
            plot_density=False,
            plot_contours=True,
            no_fill_contours=True,
            fill_contours=False,
            contour_kwargs={"zorder": 0},
        )

        plt.legend(loc=4)
        plt.xlabel(r"rise e-folding time (log day)")
        plt.ylabel(r"fade e-folding time (log day)")

        plt.xlim(0.7, 2.5)
        plt.ylim(1.0, 2.5)

        plot_fname = plot_dir + "/rise_fade.pdf"
        self.plt2box(plt, plot_fname)

        # also plot the known TDEs
        plot_fname = plot_fname.replace(".pdf", "_wtdes.pdf")
        line = plt.errorbar(
            log10(reference_tde["sigma_rise"].values) + log10(sqrt(2)),
            log10(reference_tde["sigma_fade"].values),
            color=color_dict["TDE"],
            fmt="d",
            alpha=0.7,
            lw=lw,
            mew=mew,
            ms=ms,
            zorder=1,
        )
        self.plt2box(plt, plot_fname)

        # ------
        # color / cooling
        finfo += "\n\n {0:30} {1:10} {2:10}\n".format(
            "name", "mean_color", "color_change"
        )
        ipl = (
            (results["mean_color"] != 0)
            & (np.abs(results["e_mean_color"]) < 0.35)
            & (results["n_fade"] > 3)
            & (np.abs(results["e_color_slope"]) < 0.03)
        )

        if self.plotHisto:
            plt.clf()
            plt.title("color/cooling selection; N=" + str(sum(ipl)))
            plt.hist(results[ipl]["mag_peak"], range=[17, 22], bins=20)
            plt.xlabel("peak mag")

            plot_fname = plot_dir + "/color_change_maghisto.pdf"
            self.plt2box(plt, plot_fname)

        plt.clf()
        for idx, labl, alph, fmt in zip(*plot_tuple):

            self.logger.info(f"{labl}, {sum(ipl&idx)}")
            x, y = results[ipl & idx]["mean_color"], results[ipl & idx]["color_slope"]
            x, y = x.values, y.values  # to np array

            # clip to keep very blue sources in range
            x = np.clip(x, -0.6, 0.9)

            if fmt:
                line = plt.errorbar(
                    x,
                    y,
                    xerr=results[ipl & idx]["e_mean_color"],
                    yerr=results[ipl & idx]["e_color_slope"],
                    color=color_dict[labl],
                    fmt=fmt,
                    alpha=alph,
                    lw=lw,
                    mew=mew,
                    ms=ms,
                    label=labl,
                    zorder=3 + 1 / (sum(idx) + 1),
                )

            if labl == "TDE":
                for i, nm in enumerate(results[ipl & idx]["name"]):
                    finfo += "{0:26} {1:10.3f}     {2:10.3f}\n".format(nm, x[i], y[i])
                    plt.text(x[i] - 0.01, y[i], nm, fontsize=1.5, zorder=5)
            elif (labl != "AGN") and (labl[0:2] != "SN"):
                for i, nm in enumerate(results[ipl & idx]["name"]):
                    if (
                        (y[i] < 0.02)
                        and not ("AGN?" in nm)
                        and not ("CV?" in nm)
                        and not ("varstar?" in nm)
                        or (x[i] <= 0.4)
                    ):

                        if "?" in nm:
                            fs = 1
                        else:
                            fs = 1.5
                            finfo += "{0:26} {1:10.3f}     {2:10.3f}\n".format(
                                nm, x[i], y[i]
                            )
                            plt.text(x[i], y[i], nm, fontsize=fs, zorder=5)
                        self.logger.info(f"{nm}, {not('AGN?' in nm)}")

        ax = plt.gca()
        corner.hist2d(
            reference_sne["mean_color"].values,
            reference_sne["color_slope"].values,
            bins=[10, 20],
            range=[(-0.4, 0.8), (-0.03, +0.045)],
            smooth=True,
            levels=([0.66, 0.66, 0.66]),
            ax=ax,
            color=color_dict["SN all"],
            plot_datapoints=False,
            plot_density=False,
            plot_contours=True,
            no_fill_contours=True,
            fill_contours=False,
            contour_kwargs={"zorder": 0},
        )
        plt.legend()
        plt.xlabel("mean g-r color")
        plt.ylabel("rate of color change (1/day)")
        plt.xlim(-0.6, 0.8)
        plt.ylim(-0.020, 0.04)
        # plt.draw()
        # plt.pause(0.05)

        # option to add a "TDE box"
        # plt.plot([-1,-0.00],[0.015,0.015], '--k',alpha=0.6, lw=0.5)
        # plt.plot([-0.1,-0.00],[-1,0.015], '--k',alpha=0.6, lw=0.5)
        # txt = plt.text(-0.3, -0.01, 'TDE box',  fontsize='large', verticalalignment='center', horizontalalignment='center')
        # txt.set_bbox(dict(facecolor='white', alpha=0.8, edgecolor='w'))

        plt.title("late-time (post-peak) selection")
        plot_fname = plot_dir + "/color_change.pdf"
        self.plt2box(plt, plot_fname)

        # also plot the known TDEs
        plot_fname = plot_fname.replace(".pdf", "_wtdes.pdf")
        line = plt.errorbar(
            reference_tde["mean_color"].values,
            reference_tde["color_slope"].values,
            color=color_dict["TDE"],
            fmt="d",
            alpha=0.7,
            lw=lw,
            mew=mew,
            ms=ms,
            zorder=1,
        )
        self.plt2box(plt, plot_fname)

        # ------
        # rise / color
        # print ('\n\n rise/color')
        finfo += "\n\n {0:30} {1:10} {2:10}\n".format("name", "mean_color", "rise_time")
        ipl = (
            (results["mean_color"] != 0)
            & (results["n_rise"] > 2)
            & (results["sigma_rise"] / results["e_sigma_rise"] > 1.5)
            & (abs(results["e_mean_color"]) < 0.35)
        )

        if self.plotHisto:
            plt.clf()
            plt.title("rise-time + color selection selection; N=" + str(sum(ipl)))
            plt.hist(results[ipl]["mag_peak"], range=[17, 22], bins=10)
            plt.xlabel("peak mag")

            plot_fname = plot_dir + "/rise_color_maghisto.pdf"
            self.plt2box(plt, plot_fname)

        plt.clf()
        for idx, labl, alph, fmt in zip(*plot_tuple):

            self.logger.info(f"{labl}, {sum(ipl&idx)}")

            x, y = results[ipl & idx]["mean_color"], log10(
                results[ipl & idx]["sigma_rise"]
            ) + log10(sqrt(2))
            yerr = (
                1
                / log(10)
                * results[ipl & idx]["e_sigma_rise"]
                / results[ipl & idx]["sigma_rise"]
            )
            x, y = x.values, y.values

            # clip to keep very blue sources in the plot range
            x = np.clip(x, -0.6, 0.9)

            if fmt:
                line = plt.errorbar(
                    x,
                    y,
                    xerr=results[ipl & idx]["e_mean_color"],
                    yerr=yerr,
                    color=color_dict[labl],
                    fmt=fmt,
                    alpha=alph,
                    label=labl,
                    lw=lw,
                    mew=mew,
                    ms=ms,
                    zorder=3 + 1 / (sum(idx) + 1),
                )

            if labl == "TDE":
                for i, nm in enumerate(results[ipl & idx]["name"]):
                    finfo += "{0:26} {1:10.3f}     {2:10.3f}\n".format(nm, x[i], y[i])
                    plt.text(x[i], y[i], nm, fontsize=2, zorder=5)
            elif labl != "AGN":
                for i, nm in enumerate(results[ipl & idx]["name"]):
                    if ((x[i] < 0.3 and y[i] > 0) and not ("SN" in nm)) or (
                        x[i] < -0.1 or y[i] > 0.5
                    ):

                        if ("?" in nm) or ("Ia" in nm):
                            fs = 1
                        else:
                            fs = 1.6
                        finfo += "{0:26} {1:10.3f}     {2:10.3f}\n".format(
                            nm, x[i], y[i]
                        )
                        plt.text(x[i], y[i], nm, fontsize=fs, zorder=6)

        ax = plt.gca()
        corner.hist2d(
            reference_agn["mean_color"].values,
            log10(reference_agn["sigma_rise"].values) + log10(sqrt(2)),
            bins=[30, 30],
            range=[(-0.6, 0.8), (0, log10(300))],
            smooth=True,
            levels=([0.66, 0.66, 0.66]),
            ax=ax,
            color=color_dict["AGN"],
            plot_datapoints=False,
            plot_density=False,
            plot_contours=True,
            no_fill_contours=True,
            fill_contours=False,
            contour_kwargs={"zorder": 0},
        )  # 'lw':1 keyword not used by corner
        corner.hist2d(
            reference_sne["mean_color"].values,
            log10(reference_sne["sigma_rise"].values) + log10(sqrt(2)),
            bins=[30, 30],
            range=[(-0.6, 0.8), (0, log10(300))],
            smooth=True,
            levels=([0.66, 0.66, 0.66]),
            ax=ax,
            color=color_dict["SN all"],
            plot_datapoints=False,
            plot_density=False,
            plot_contours=True,
            no_fill_contours=True,
            fill_contours=False,
            contour_kwargs={"zorder": 0},
        )

        plt.title("real-time (pre-peak) selection")
        plt.legend(loc=1)
        plt.xlabel("mean g-r color")
        plt.ylabel(r"rise e-folding time (log$_{10}$ day)")
        plt.xlim(-0.6, 0.8)
        plt.ylim(log10(4), log10(300))

        plot_fname = plot_dir + "/rise_color.pdf"
        self.plt2box(plt, plot_fname)

        # also plot the known TDEs
        plot_fname = plot_fname.replace(".pdf", "_wtdes.pdf")
        line = plt.errorbar(
            reference_tde["mean_color"].values,
            log10(reference_tde["sigma_rise"].values) + log10(sqrt(2)),
            color=color_dict["TDE"],
            fmt="d",
            alpha=0.7,
            lw=lw,
            mew=mew,
            ms=ms,
            zorder=1,
        )
        self.plt2box(plt, plot_fname)

        # as a bonus: write a file with the names of TDE candidates (based on some boxing)
        # for each of the plots (requested by Suvi)
        self.put(plot_dir + "/boxed.txt", bytes(finfo, "utf-8"))

        self.commit()

    def collect_metrics(self, transients):
        """ """
        jd_today = self.night.jd

        alert_name = "/mampel/alerts"
        alert_folders = self.get_files(alert_name)
        if "ps1_offset.json" not in alert_folders:
            buf = io.BytesIO()
            buf.write(bytes(json.dumps({}), "utf-8"))
            buf.seek(0)
            self.put(alert_name + "/ps1_offset.json", buf.read())
            ps1_offset = {}
        else:
            ps1_offset = self.read_file(alert_name + "/ps1_offset.json")[1].json()

        if transients is not None:

            local_sources = np.array(transients)  # deprecated daysback soln
            newdetection_count = len(local_sources)
            self.logger.info(
                "Collecting metrics for {} transient(s)".format(newdetection_count)
            )
            metrics, metrics_flex = [], []
            for i, tran_view in enumerate(local_sources):

                # get classifcation plus some neoWISE info
                classification, extra_info = classifyme.get_class(
                    tran_view, self, self.logger, write=True
                )

                metrics_flex.append(self.get_fit_params(tran_view, classification))
                metrics.append(self.get_simple_results(tran_view, classification))

            metrics_flex = pd.DataFrame(metrics_flex, columns=flex_keys)
            metrics = pd.DataFrame(metrics, columns=simple_keys)

            self.logger.info(
                "successfully collected metrics: "
                + str(metrics_flex["name"].all() == metrics["name"].all())
            )
            if metrics_flex["name"].all() != metrics["name"].all():
                warnings.warn(
                    "WARNING: flex metrics and simple metrics are not aligned, plots may be erroneous."
                )
        else:
            self.logger.info("no transients to rank!")

        return metrics, metrics_flex

    def get_fit_params(self, tran_view, classification):
        try:
            fit_params = (tran_view.get_t2_result("T2FlexFit") or {})["fit_params"]
            return {**fit_params, **{"classification": classification}}

        except KeyError:
            self.logger.warn(
                "WARNING: No Flex fit for source {}".format(tran_view.stock["name"][0])
            )
            return {
                "name": tran_view.stock["name"][0],
                "classification": classification,
            }

    def get_simple_results(self, tran_view, classification):
        try:
            simple_results = (tran_view.get_t2_result("T2SimpleMetrics") or {})[
                "metrics"
            ]
            return {**simple_results, **{"classification": classification}}

        except KeyError:
            self.logger.warn(
                "WARNING: No simple metrics for source {}".format(
                    tran_view.stock["name"][0]
                )
            )
            return {
                "name": tran_view.stock["name"][0],
                "classification": classification,
            }

    def datetomjd(self, d):
        d0 = datetime.datetime(1858, 11, 17, 0, 0, 0)
        dt = d - d0
        # dt is a timedelta object.
        return dt.days + (dt.seconds + dt.microseconds / 1e6) / 86400.0

    def mjdtodate(self, mjd):
        if not np.isfinite(mjd):
            return None
        jd = mjd + 2400000.5
        unixtime = (jd - 2440587.5) * 86400.0  # in seconds
        return datetime.datetime.utcfromtimestamp(unixtime)

    def format_date(self, transient_date):
        return astropy.time.Time(
            transient_date[6:10]
            + "-"
            + transient_date[3:5]
            + "-"
            + transient_date[0:2]
            + transient_date[10:]
        ).jd



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os, sys, pickle, json, datetime

import pytest

from ampel.log.AmpelLogger import AmpelLogger
from ampel.secret.AmpelVault import AmpelVault
from ampel.secret.DictSecretProvider import DictSecretProvider

logger = AmpelLogger.get_logger()

now = datetime.datetime.now()


@pytest.fixture
def single_TransientView_fixture():
    with open(
        os.path.join(os.path.dirname(__file__), "ZTF20aaaaflr_transientview.pickle"),
        "rb",
    ) as f:
        tview = pickle.load(f)
    return tview


@pytest.fixture
def multi_TransientView_fixture():
    with open(
        os.path.join(os.path.dirname(__file__), "transientview_list.pickle"), "rb"
    ) as f:
        tview_list = pickle.load(f)
    return tview_list


@pytest.fixture
def dropbox_token():
    from ampel.secret.NamedSecret import NamedSecret

    return NamedSecret(label="dropbox/token", value=os.environ["DROPBOX_TOKEN"])


@pytest.fixture
def testdir():
    midnight = now.replace(hour=0, minute=0, second=0, microsecond=0)
    seconds_to_midnight = (midnight - now).seconds

    out_dict = {
        "testdir": now.strftime("/mampel/test/%Y-%m-%d_%H%M"),
        "force_date": now.strftime("%Y-%m-%d"),
        "force_year": now.strftime("%Y"),
        "force_md": now.strftime("%m-%d"),
        "datetime": now,
    }

    if seconds_to_midnight / 60 < 10:
        print(
            "Be wary: we are within {0:0.1f} minutes of midnight; testing runs for fixed date: {1}".format(
                seconds_to_midnight60, out_dict["force_date"]
            )
        )

    return out_dict


@pytest.fixture
def true_output():
    # In case we want to add other comparison files in the future, unused currently
    truth = {}
    with open("ampel/nuclear/test/ZTF20aaaaflr_wisedump.json") as f:
        truth["wise"] = json.loads(f.read())

    return truth


def test_t3_ranking(multi_TransientView_fixture, dropbox_token, testdir):

    from ampel.nuclear.t3.T3Ranking import T3Ranking

    # don't run this right before midnight :)
    # path = "{0}/ranking/{1}/{2}_everything.txt".format(
    #     testdir["testdir"], testdir["force_year"], testdir["force_md"]
    # )
    path = f"{testdir['testdir']}/ranking/{testdir['force_year']}/{testdir['force_md']}_everything.txt"

    print("test_t3_ranking: creating ranking files in:\n{0}".format(path.split("_")[0]))

    sum_dir = (
        f"{testdir['testdir']}/sum_plots/{testdir['force_year']}/{testdir['force_md']}"
    )

    print("test_t3_ranking: creating summary plots in:\n{0}".format(sum_dir))

    Rank_T3 = T3Ranking(
        logger=logger,
        dryRun=False,
        base_location=testdir["testdir"],
        dropbox_token=dropbox_token,
    )
    Rank_T3.post_init()

    # test running with no input
    Rank_T3.process([])

    # need to init again
    Rank_T3.post_init()

    # test running all transients
    Rank_T3.process(multi_TransientView_fixture)

    # check that ranking and summary plot files are created
    print(Rank_T3.stats["files"])
    assert Rank_T3.stats["files"] > 6  # check all of the files were created

    # check for ranking
    nlines = Rank_T3.read_file(path).text.count("ZTF")
    print(Rank_T3.read_file(path).text)
    print(nlines)
    assert nlines > 45  # check number of lines in ranking, at least 46 lines long

    # check for summary
    assert Rank_T3.exists(sum_dir + "/rise_color.pdf")
    assert Rank_T3.exists(sum_dir + "/rise_fade.pdf")
    assert Rank_T3.exists(sum_dir + "/color_change.pdf")


def test_t3_metricsplots(single_TransientView_fixture, dropbox_token, testdir):

    from ampel.nuclear.t3.T3MetricsPlots import T3MetricsPlots

    # Test MetricsPlots
    Metrics_T3 = T3MetricsPlots(
        logger=logger,
        verbose=True,
        dryRun=False,
        base_location=testdir["testdir"],
        dropbox_token=dropbox_token,
    )
    Metrics_T3.post_init()

    path = f"{testdir['testdir']}/alerts/2020/ZTF20aaaaflr/ZTF20aaaaflr_flex.json"

    print("test_t3_metricsplots: creating single flexfit output at\n{0}".format(path))

    Metrics_T3.process([single_TransientView_fixture])

    metrics_dump = Metrics_T3.read_file(path).json()
    assert metrics_dump["name"] == "ZTF20aaaaflr"
    assert Metrics_T3.stats["files"] == 5  # ensure we save every file properly


def test_t3_plot_neowise(single_TransientView_fixture, dropbox_token, testdir):

    from ampel.nuclear.t3.T3PlotNeoWISE import T3PlotNeoWISE

    # Test PlotNeoWISE
    WISE_T3 = T3PlotNeoWISE(
        logger=logger,
        verbose=True,
        base_location=testdir["testdir"],
        apply_qcuts=True,
        plot_allWISE=False,
        dryRun=False,
        dropbox_token=dropbox_token,
    )
    WISE_T3.post_init()

    path = f"{testdir['testdir']}/alerts/2020/ZTF20aaaaflr/ZTF20aaaaflr_neoWISE.json"

    print("test_t3_plot_neowise: creating new neoWISE output in:\n{0}".format(path))

    WISE_T3.process([single_TransientView_fixture])

    wise_dump = WISE_T3.read_file(path).json()
    assert WISE_T3.stats["files"] >= 3
    assert list(wise_dump.keys()) == ["info_list", "wise_class", "out_dict"]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ampel',
 'ampel.nuclear.flexfit',
 'ampel.nuclear.t0',
 'ampel.nuclear.t2',
 'ampel.nuclear.t3',
 'ampel.nuclear.test']

package_data = \
{'': ['*']}

install_requires = \
['ampel-hu-astro[slack,sncosmo]>=0.8.3a2,<0.8.4',
 'ampel-ztf>=0.8.3a7,<0.8.4',
 'corner>=2.2.1,<3.0.0',
 'dropbox>=11.36.0,<12.0.0']

setup_kwargs = {
    'name': 'ampel-nuclear',
    'version': '0.8.3a4',
    'description': 'Astronomy units for the Ampel system to analyze nuclear events like TDEs',
    'long_description': "# Ampel-nuclear\nCentral repository to host AMPEL code to search for and analyze nuclear transients. At the moment, this is exclusively code from the ZTFbh science working group.\n\n## Installation\n### Prerequisites\nYou need to export environment variables for the [AMPEL ZTF archive](https://ampelproject.github.io/astronomy/ztf/index) (tokens are available [here](https://ampel.zeuthen.desy.de/live/dashboard/tokens)), for [Fritz](https://fritz.science/), and for the dropbox API (ask Sjoert). \n\nFurthermore, you need a running instance of [MongoDB](https://www.mongodb.com/docs/manual/installation/).\n\n### Setup\nCreate a fresh Python 3.10 conda env\n```\nconda create -n tde_filter_upgrade python=3.10\nconda activate tde_filter_upgrade\n```\nInstall is done via poetry:\n```\npip install poetry \ngit clone https://github.com/AmpelProject/ampel-nuclear\ncd Ampel-nuclear\npoetry install\n```\nNow we have to build the ampel config. Issue\n```\nampel config build -out ampel_conf.yaml\n```\nNote: this will throw some import errors, but you can ignore these because those packages are not needed locally. \n\nNow you need to export the following tokens\n```\nexport AMPEL_ARCHIVE_TOKEN='' \nexport DROPBOX_TOKEN=''\nexport FRITZ_TOKEN=''\n```\n\n## Test\nTo run the test, start your local MongoDB. And then issue\n\n```\n./run_tde_scan.py -i\n```\n\nThe `-i` initiates (and saves) a new archive API stream token. To change the date, use `-d YYYY-MM-DD` for a certain day. The script will request alerts for the 24 hours after this date. \n\nNote: When requesting a full day with `-d` from the archive, the first run will probably fail, as the archive database has not fully ramped up yet (`URL is locked`). In this case, just rerun `./run_tde_scan.py -d YYYY-MM-DD` (without `i` to prevent requesting a new stream token and overwriting the current one) until the archive starts serving alerts (you will see them getting ingested).\n\nTo check the output, go to the `temp` directory that gets created when script is run without `-p` (push to dropbox).\n\nTo see all available commands of the test script, run `./run_tde_scan.py -h`.\n",
    'author': 'Valery Brinnel',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://ampelproject.github.io',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<3.11',
}


setup(**setup_kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : ampel/nuclear/t3/dropboxIO.py
# License           : BSD-3-Clause
# Author            : ?
# Date              : ?
# Last Modified Date: 29.11.2022
# Last Modified By  : simeon.reusch@desy.de

import dropbox  # type: ignore
import os
import tempfile
from typing import Optional, Union
from collections.abc import Generator
from functools import lru_cache
import datetime
from concurrent.futures import ThreadPoolExecutor

import astropy  # type: ignore
import astropy.time  # type: ignore
from astropy import units as u
import backoff
from requests.exceptions import ConnectionError

from ampel.types import UBson, T3Send
from ampel.view.TransientView import TransientView
from ampel.view.T3Store import T3Store
from ampel.struct.UnitResult import UnitResult
from ampel.abstract.AbsPhotoT3Unit import AbsPhotoT3Unit
from ampel.secret.NamedSecret import NamedSecret

from ampel.log.AmpelLogger import AmpelLogger

handle_disconnects = backoff.on_exception(backoff.expo, ConnectionError, max_time=300)


class DropboxUnit(AbsPhotoT3Unit):

    dropbox_token: NamedSecret[str]
    dryRun: bool = False  #: whether to write (and sometimes read) from the dropbox
    dryRunDir: Optional[
        str
    ] = None  # if dryRun is enabled, you must specify a desired temporary directory
    base_location: str = "/mampel"  #:optional diff. directory for testing
    max_connections: int = 20
    date: Optional[str] = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        dbx = None
        if self.dryRun:
            if not self.dryRunDir:
                raise ValueError("You have to specify dryRunDir when using dryRun")

        dbx = dropbox.Dropbox(
            self.dropbox_token.get(),
            session=dropbox.create_session(max_connections=self.max_connections + 1),
        )
        try:
            acct = dbx.users_get_current_account()

            if (acct.email == "mitchell.karmen@nyu.edu") or (
                acct.email == "sjoertvanvelzen@gmail.com"
            ):
                self.dbx = dbx
            else:
                raise ValueError(
                    f"Invalid dropbox account {acct.email}. Please check access token."
                )
        except Exception as exc:
            raise ValueError(
                "Invalid access token. Please check login credentials."
            ) from exc

        if self.dryRun:
            if self.dryRunDir:
                if not os.path.exists(self.dryRunDir):
                    os.makedirs(self.dryRunDir)
                self.tmpdir = self.dryRunDir
            else:
                self.tmpdir = tempfile.mkdtemp()
            self.logger.info(
                f"Running in dryRun mode, saving to temporary directories. These are here: {self.tmpdir}"
            )

        self.stats = {"bytes": 0, "files": 0}
        self._uploads = {}
        self._payloads = []

        if self.base_location == "/mampel":
            if self.date is not None:
                date_format = "%Y-%m-%d"
                req_date = str(datetime.datetime.strptime(self.date, date_format))
                self.night = (
                    astropy.time.Time(req_date, format="iso", scale="utc")
                    + 0.99 * u.day
                )
            else:
                self.night = astropy.time.Time.now()
        else:
            self.night = astropy.time.Time(
                self.base_location[13:23]
            )  # shouldn't change from test to test
            if self.dryRun:
                self.create_folder(self.base_location)
            if self.base_location[-15:] not in self.get_files(
                "/mampel/test"
            ):  # hard coding in test dir, 'exists' function on works for files
                self.create_folder(self.base_location)

    def process(
        self, gen: Generator[TransientView, T3Send, None], t3s: Optional[T3Store] = None
    ) -> Union[UBson, UnitResult]:
        """ """
        return None

    def done(self):
        self.commit()
        self.logger.info(
            f"Wrote {self.stats['bytes']/2**20:.3f} MB in {self.stats['files']} files"
        )

    @handle_disconnects
    def create_folder(self, path):
        if self.dryRun:
            if path.startswith("/"):
                path = path[1:]
            path = os.path.join(self.tmpdir, path)
            os.makedirs(path, exist_ok=True)
            self.logger.info(f"Create {path}")
        else:
            try:
                self.dbx.files_create_folder_v2(path)
            except dropbox.exceptions.ApiError as exc:
                # ignore conflict with existing folder
                if (
                    isinstance(exc.error, dropbox.files.CreateFolderError)
                    and exc.error.is_path()
                    and exc.error.get_path().is_conflict()
                    and exc.error.get_path().get_conflict().is_folder()
                ):
                    ...
                else:
                    raise

    @handle_disconnects
    def put(self, path, payload):
        if self.dryRun:
            if path.startswith("/"):
                path = path[1:]
            path = os.path.join(self.tmpdir, path)
            if not os.path.isdir(os.path.dirname(path)):
                os.makedirs(os.path.dirname(path), exist_ok=True)
            with open(path, "wb") as f:
                f.write(payload)
            self.logger.info(f"Wrote {len(payload)} bytes to {path}")
        else:
            self._payloads.append((payload, path, len(payload)))
        self.stats["bytes"] += len(payload)
        self.stats["files"] += 1

    def commit(self) -> None:

        # Dropbox cannot handle queues containing more than 1000 files
        payload_subsets = [
            self._payloads[i : i + 1000] for i in range(0, len(self._payloads), 1000)
        ]

        # Iterate over payload chunks, create a new pool for each
        for payload_subset in payload_subsets:
            entries = []
            pool = ThreadPoolExecutor(max_workers=self.max_connections)
            futures = {}

            # Fill the pool
            for payload, path, offset in payload_subset:
                f = pool.submit(
                    handle_disconnects(self.dbx.files_upload_session_start),
                    payload,
                    close=True,
                )
                futures[f] = (path, len(payload))

            for future, (path, offset) in futures.items():

                start_result = future.result()
                entries.append(
                    dropbox.files.UploadSessionFinishArg(
                        cursor=dropbox.files.UploadSessionCursor(
                            session_id=start_result.session_id, offset=offset
                        ),
                        commit=dropbox.files.CommitInfo(
                            path=path, mode=dropbox.files.WriteMode.overwrite
                        ),
                    )
                )
            if not entries:
                return

            self.logger.info(f"committing {len(entries)} uploads")
            launch_result = handle_disconnects(
                self.dbx.files_upload_session_finish_batch
            )(entries)
            if not launch_result.is_complete():
                status = backoff.on_predicate(
                    backoff.expo,
                    predicate=lambda job_status: job_status.is_in_progress(),
                    max_time=300,
                    max_value=5,
                )(handle_disconnects(self.dbx.files_upload_session_finish_batch_check))(
                    launch_result.get_async_job_id()
                )
                assert status.is_complete()
                for future, result in zip(
                    list(futures.keys()), status.get_complete().entries
                ):
                    if result.is_success():
                        del futures[future]
                    else:
                        raise RuntimeError(str(result.get_failure()))

    @lru_cache(maxsize=1024)
    @handle_disconnects
    def get_files(self, path):
        if not self.dbx:
            self.dbx = dbx
        entries = []
        if self.dryRun:
            if path.startswith("/"):
                temp_path = path[1:]
            temp_path = os.path.join(self.tmpdir, temp_path)
            try:
                entries = os.listdir(temp_path)
                return entries
            except Exception as e:
                self.logger.info("Directory not found locally, checking dropbox...")

        try:
            list_folder = self.dbx.files_list_folder(path)
        except dropbox.exceptions.ApiError as exc:
            if (
                isinstance(exc.error, dropbox.files.ListFolderError)
                and exc.error.is_path()
                and exc.error.get_path().is_not_folder()
            ):
                raise NotADirectoryError(f"{path} is not a directory") from exc
            else:
                print(path)
                raise

        entries += [entry.name for entry in list_folder.entries]
        has_more = list_folder.has_more
        while has_more:
            list_folder = self.dbx.files_list_folder_continue(list_folder.cursor)
            entries += [entry.name for entry in list_folder.entries]
            has_more = list_folder.has_more

        return entries

    @lru_cache(maxsize=1024)
    @handle_disconnects
    def exists(self, path):
        try:
            self.dbx.files_get_metadata(path)
            return True
        except:
            return False

    @handle_disconnects
    def read_file(self, path):
        if self.dryRun:
            filepath_local = os.path.join(
                self.tmpdir, path[1:] if path.startswith("/") else path
            )

            if os.path.exists(filepath_local):
                with open(filepath_local) as f:
                    return f.read()
            else:
                self.logger.info(
                    "File {} not found locally, searching dropbox".format(
                        filepath_local
                    )
                )

        try:
            metadata, f = self.dbx.files_download(path)
            return f
        except dropbox.exceptions.ApiError as exc:
            if (
                isinstance(exc.error, dropbox.files.DownloadError)
                and exc.error.is_path()
                and exc.error.get_path().is_not_found()
            ):
                raise FileNotFoundError(f"{path} is not a file") from exc
            else:
                raise
